var LEFT_MOUSE_BUTTON = 0;
var MIDDLE_MOUSE_BUTTON = 1;

var KEY_A = 65;
var KEY_ESC = 27;

var Graph = function(domPath, nodes, edges, browser) {
	var self = this;

	self.graph;
	self.graphics;
	self.renderer;
	self.element; // jQuery DOM element

	var dragging = {
		node: null,
		pos: []
	}

	var popup = undefined;

	var selection = {
		on: 'false',
		start: [], // [x, y]
		end: [], // [x, y]
		nodes: [],
		element: undefined,
		overlay: undefined
	}

	self.isBrowser = false;

	/*self.anchorPosition = {
		// <nodeId>: { <edgeId>: { x: ..., y: ..., pinned: <bool> }}
	}

	self.linking = {
		link: null,
		mode: undefined, // start/end
		pos: {}
	}*/

	// Constructor.
	// Do NOT call this function.
	var initialize = function(domPath, nodes, edges, browser) {
		self.isBrowser = browser === true;
		self.element = $(domPath);
		selection.element = $(self.element).children('.selection');
		selection.overlay = $(self.element).next('#graph-overlay');

		// Initialize engines
		self.graph = Viva.Graph.graph();
		self.graphics = Viva.Graph.View.svgGraphics();
		self.layout = Viva.Graph.Layout.forceDirected(self.graph, {
		    springLength: 250,
		    springCoeff: 0.0008,
		    dragCoeff: 0.002,
		    gravity: -2,
		    theta: 1
	  	});

	  	self.renderer = Viva.Graph.View.renderer(self.graph, {
        	layout: self.layout,
        	graphics: self.graphics,
        	container: self.element[0],
        	prerender: true
        });

	  	// Set-up elements
	  	nodes.forEach(function(element) {
	  		self.addNode(element.id, element.data);
	  	});

	  	edges.forEach(function(element) {
	  		self.addEdge(element.fromId, element.toId, element.data);
	  	});

	    // Customize renderer
	  	self.graphics.getSvgRoot().attr('version', '1.1'); // avoid winforms errors
	  	//graphics.getSvgRoot().setAttribute('xmlns', 'http://www.w3.org/2000/svg');
	  	//graphics.getSvgRoot().setAttribute('xmlns:xlink', 'http://www.w3.org/1999/xlink');

	  	self.graphics.node(getNodeSvg);
	  	self.graphics.link(getEdgeSvg);
	  	self.graphics.placeNode(renderNode);
	  	self.graphics.placeLink(renderEdge);

	  	// Default layout
		initClipPaths();
	  	self.arrangeNodes();
	  	self.pinNodes();

	  	self.renderer.run();
	  	
	  	if(!self.isBrowser) {
		  	bindGraphEvents();
		}
	}

	// Common used svg clip-paths (shapes).
	var initClipPaths = function() {
		// Entity ClipPath.
		// Crop text going outside rect.
		var entityClipPath = Viva.Graph.svg('clipPath')
			.attr('id', 'entity');

		var rect = Viva.Graph.svg('rect')
			.attr('width', self.entity.width - 2*self.entity.padding)
			.attr('height', self.entity.height)
			.attr('x', self.entity.padding);

		entityClipPath.append(rect);
		self.graphics.getSvgRoot().append(entityClipPath);
	}

	var highlightNode = function(node) {
		if(dragging.node == null) {
			self.graph.forEachNode( function(node) {
				self.graphics.getNodeUI(node.id).attr('highlighted', 'false');
			});

			var links = self.graph.getLinks(node.id);
			links.forEach( function(link) {
				var A = self.graphics.getNodeUI(link.fromId);
				var B = self.graphics.getNodeUI(link.toId);

				A.attr('highlighted', 'half');
				B.attr('highlighted', 'half');
			});

			var ui = self.graphics.getNodeUI(node.id).attr('highlighted', 'true');
		}
	}

	var removeNodeHighlight = function(node) {
		if(dragging.node == null) {
			self.graph.forEachNode( function(node) {
				self.graphics.getNodeUI(node.id).attr('highlighted', 'half');
			});
		}
	}

	var highlightEdge = function(edge) {
		var ui = self.graphics.getLinkUI(edge.id);

		ui.childNodes[0]
			.attr('stroke', self.entity.hover.backgroundColor);
	}

	var removeEdgeHighlight = function(edge) {
		var ui = self.graphics.getLinkUI(edge.id);

		ui.childNodes[0]
			.attr('stroke', 'transparent');
	}

	var startSelection = function(x, y) {
		selection.on = 'true';
		selection.start = [x, y];
		updateSelection(x, y);
		
		selection.nodes.forEach(function(node) {
			setNodeUnselected(node.id);
		});

		selection.nodes = [];
		$(selection.element).show();
	}

	var updateSelection = function(x, y) {
		selection.end = [x, y];

		// update 
		var r = normalizeSelection(selection.start, selection.end);
		$(selection.element).css({
			left: r.x +'px',
			top: r.y +'px',
			width: r.width +'px',
			height: r.height +'px'
		});
	}

	// @param a, b Positions of opposite corners of the rectangle in [x, y] format.
	var normalizeSelection = function(a, b) {
		return {
			x: Math.min(a[0], b[0]),
			y: Math.min(a[1], b[1]),
			width: Math.abs(a[0] - b[0]),
			height: Math.abs(a[1] - b[1])
		}
	}

	var endSelection = function() {
		updatesNodesInSelection(selection.start, selection.end);
		selection.on = 'false';

		$(selection.element).hide();
	}

	var screenToGraphOffset = function() {
		var g = self.graphics.getSvgRoot().childNodes[0];
		var mat = g.attr('transform').match(/matrix\((.*)\)/)[1];
		var splitChar = mat.indexOf(',') >= 0 ? ',' : ' ';
		var offset = mat.split(splitChar);

		return {
			x: -offset[4],
			y: -offset[5]
		};
	}

	var screenToGraphPosition = function(pos) {
		var offset = screenToGraphOffset();

		return {
			x: pos.x + offset.x,
			y: pos.y + offset.y
		}
	}

	var updatesNodesInSelection = function(a, b) {
		var r = normalizeSelection(a, b);

		// transform screen coordinate to graph coordinate
		//var offset = layout.getGraphRect();
		var offset = screenToGraphOffset();

		r.x += offset.x;
		r.y += offset.y;

		// store selected
		self.graph.forEachNode( function(node) {
			var nodePos = self.layout.getNodePosition(node.id);
			if(nodePos.x - self.entity.width/2 >= r.x
			&& nodePos.y - self.entity.height/2 >= r.y
			&& nodePos.x + self.entity.width/2 <= r.x + r.width
			&& nodePos.y + self.entity.height/2 <= r.y + r.height) {
				selection.nodes.push(node);
				setNodeSelected(node.id);
			}
		});
	}

	var setNodeSelected = function(nodeId) {
		self.graphics.getNodeUI(nodeId).attr('selected', 'true');
		updateNodePosProperty(self.graph.getNode(nodeId));
	}

	// Store current position, used to verify if position changed
	// (avoid sending unwanted notifications to C# control).
	var updateNodePosProperty = function(node) {
		var pos = self.layout.getNodePosition(node.id);
		node.data.pos = [pos.x, pos.y];
	}

	var isNodeSelected = function(nodeId) {
		return self.graphics.getNodeUI(nodeId).attr('selected') === 'true';
	}

	var setNodeUnselected = function(nodeId) {
		self.graphics.getNodeUI(nodeId).attr('selected', 'false');
		self.renderer.resume();
	}

	var bindGraphEvents = function() {
		$(document).keydown(function(e) {
			// Ctrl + A
	        if(e.which == KEY_A && e.ctrlKey) {
	            if(selection.nodes.length != self.graph.getNodesCount()) {
	        		// Select all if at least one is not selected
	            	self.graph.forEachNode(function(node) {
	            		if(selection.nodes.indexOf(node) < 0) {
			            	selection.nodes.push(node);
							setNodeSelected(node.id);
						}
					});
	            }
	            else {
	            	// Cancel selection if all selected.
	            	selection.nodes.forEach(function(node) {
						setNodeUnselected(node.id);
					});
					selection.nodes = [];
	            }
	        }

	        // Esc - cancel selection
	        if(e.which == KEY_ESC) {
	        	selection.nodes.forEach(function(node) {
					setNodeUnselected(node.id);
				});
				selection.nodes = [];
	        }

	        // Disable all other keys
	        e.preventDefault();
	        e.stopPropagation();
	        e.stopImmediatePropagation();
	    });

		$(self.element).children('svg').mousedown(function(e) {
			if(e.ctrlKey) {
				e.preventDefault();
				e.stopPropagation();
				return;
			}

			if(e.button == LEFT_MOUSE_BUTTON) {
				e.preventDefault();
				e.stopPropagation();

				selection.overlay.show();
				startSelection(e.clientX, e.clientY);
			}
		});

		$(window.document).mousemove(function(e) {
			if(selection.on == 'true') {
				e.preventDefault();
				e.stopPropagation();

				updateSelection(e.clientX, e.clientY);
			}
		}).mouseup(function(e) {
			if(selection.on == 'true') {
				endSelection();
				selection.overlay.hide();
			}

			if(e.ctrlKey) {
				selection.overlay.hide();
			}
		});

		$(window).mousemove(function(e){
			// Change selected nodes position
			if(dragging.node != null) {
				var newPos = self.layout.getNodePosition(dragging.node.id);
				selection.nodes.forEach(function(node) {
					if(node != dragging.node) {
						var pos = self.layout.getNodePosition(node.id);
						self.layout.setNodePosition(node.id,
							pos.x + newPos.x - dragging.pos.x,
							pos.y + newPos.y - dragging.pos.y
						);
					}
				});

				dragging.pos = {
					x: newPos.x,
					y: newPos.y
				};
			}
		});
	}

	var bindNodeBrowserEvents = function(node, domElement) {
		var ui = $(domElement);

		ui.mousedown(function(e) {
			e.stopImmediatePropagation();
		});
	}

	var bindNodeEvents = function(node, domElement) {
		var ui = domElement;

		$(ui).mousedown(function(e) {
			// start dragging
			dragging.node = node;
			dragging.pos = {
				x: self.layout.getNodePosition(node.id).x,
				y: self.layout.getNodePosition(node.id).y
			};

			if(!e.ctrlKey) {
				// cancel selection
				if(selection.nodes.indexOf(dragging.node) < 0) {
					selection.nodes.forEach(function(node) {
						setNodeUnselected(node.id);
					});
					selection.nodes = [];

					// and select current
					selection.nodes.push(node);
					setNodeSelected(node.id);
				}
			}

			if(e.ctrlKey) {
				e.stopImmediatePropagation();
				selection.overlay.show();

				var index = selection.nodes.indexOf(node);
				if(index < 0) {
					selection.nodes.push(node);
					setNodeSelected(node.id);
				}
				else {
					setNodeUnselected(node.id);
					selection.nodes.splice( $.inArray(node, selection.nodes), 1 );
				}
			}
		});

		$(ui).mouseup(function(e) {
			// notify node position change
			// bug: called also when pos not change (eq. click)
			selection.nodes.forEach(function(node) {
				var pos = self.layout.getNodePosition(node.id);

				if(parseInt(pos.x) != parseInt(node.data.pos[0])
				|| parseInt(pos.y) != parseInt(node.data.pos[1])) {
					updateNodePosProperty(node);
					self.onNodeChanged(node);
				}
			})
		});

		$(document).mouseup(function() {
			// stopped dragging, cannot be catched on ui
			// (can stop dragging when mouse is NOT over the element).
			dragging.node = null;
		});

		$(ui).hover(function() {
			// mouse enter - highlight node and neighbours
			highlightNode(node);
		}, function() {
			// mouse leave - remove higlight
			removeNodeHighlight(node);
		});

		// Double click
		$(ui).dblclick(function() {
			if(window.external == 'Dataedo.Tools.ERD.ERDEditor') {
				window.external.NotifyNodeDblClick(node.id);
			}
		});
	}

	var bindEdgeEvents = function(edge, domElement) {
		var ui = domElement;

		$(ui).hover(function() {
			// mouse enter - highlight node and neighbours
			highlightEdge(edge);
		}, function() {
			// mouse leave - remove higlight
			removeEdgeHighlight(edge);
		});

		/*$(ui).mousedown(function(e) {
			e.stopImmediatePropagation();

			// TODO: ...
			setClosestNodeLinkAnchor(edge, {
				x: e.clientX,
				y: e.clientY
			});
		});*/
	}

	/*var setClosestNodeLinkAnchor = function(edge, cursorPos) {
		var fromPos = self.layout.getNodePosition(edge.fromId);
		var toPos = self.layout.getNodePosition(edge.toId);

		var anchors = getEdgePoints(fromPos, toPos);
		var offset = screenToGraphOffset();

		var last = anchors.length-1;
		anchors[0].x -= offset.x;
		anchors[0].y -= offset.y;
		anchors[last].x -= offset.x;
		anchors[last].y -= offset.y;

		var fromDistance = vectorDistance(cursorPos, anchors[0]);
		var toDistance = vectorDistance(cursorPos, anchors[last]);

		self.linking.link = edge;
		self.linking.mode = (fromDistance <= toDistance) ? 'start' : 'end';
		self.linking.pos = cursorPos;

		var id = self.linking.mode == 'start' ? edge.toId : edge.fromId;
		var details = (self.linking.mode == 'start') ? getNodePosDetails(toPos, fromPos) : getNodePosDetails(fromPos, toPos);
		self.anchorPosition[id] = self.anchorPosition[id] || {
			x: (self.linking.mode == 'start') ? (anchors[last].x + offset.x) : (anchors[0].x + offset.x),
			y: (self.linking.mode == 'start') ? (anchors[last].y + offset.y) : (anchors[0].y + offset.y),
		};
	}*/

	var vectorDistance = function(a, b) {
		var x2 = (b.x-a.x) * (b.x-a.x);
		var y2 = (b.y-a.y) * (b.y-a.y);

		return Math.sqrt(x2 + y2);
	}

	var updateNodeIcon = function(node) {
		var nodeUI = self.graphics.getNodeUI(node.id);

		var srcSet = node.data.deleted === true ? self.entity.icon.deletedNode : self.entity.icon.node;
		nodeUI.childNodes[1].link(srcSet[node.data.type]);
	}

	// Catch nodes properties changes incoming from C#.
	var notifyNodePropertyChange = function(nodeId, propertyName, value) {
		var node = self.graph.getNode(nodeId);
		var nodeUI = self.graphics.getNodeUI(nodeId);

		// Node not found.
		if(node == undefined)
			return;

		switch(propertyName) {
			case 'Deleted':
				node.data.deleted = value === 'True';
				updateNodeIcon(node);
				break;

			case 'Type':
				node.data.type = value.toLowerCase();
				updateNodeIcon(node);
				break;

			case 'Label':
				node.data.label = value;
				nodeUI.childNodes[2].text(node.data.label);
				break;

			case 'Popup':
				node.data.popup = value;
				$(nodeUI).popup({ msg: node.data.popup });
				break;

			case 'PosX':
				var x = parseInt(value);
				self.setNodePos(nodeId, x, null);
				break;

			case 'PosY':
				var y = parseInt(value);
				self.setNodePos(nodeId, null, y);
				break;
		}
	}

	var getNodeSvg = function(node) {
		var ui = Viva.Graph.svg('g');

		if(node.data.popup != undefined)
			$(ui).popup({ msg: node.data.popup });
		
		var rect = Viva.Graph.svg('rect')
			.attr('width', self.entity.width)
			.attr('height', self.entity.height)
			.attr('fill', self.entity.normal.backgroundColor)
			.attr('stroke', self.entity.normal.borderColor)
			.attr('stroke-width', 1);

		var srcSet = node.data.deleted === true ? self.entity.icon.deletedNode : self.entity.icon.node;
		var image = Viva.Graph.svg('image')
			.link(srcSet[node.data.type])
			.attr('width', self.entity.icon.width)
			.attr('height', self.entity.icon.height)
			.attr('x', self.entity.padding)
			.attr('y', self.entity.height/2 - self.entity.icon.height/2);

		var label = Viva.Graph.svg('text')
			.attr('text-anchor', 'left')
			.attr('clip-path', 'url(#entity)')
			.attr('font-family', 'Arial')
			.attr('font-size', '12')
			.attr('dy', '5px')
			.attr('x', 1.5*self.entity.padding + self.entity.icon.width)
			.attr('y', self.entity.height / 2)
			.text(node.data.label);

		ui.append(rect);
		ui.append(image);
		ui.append(label);

		if(!self.isBrowser) {
			bindNodeEvents(node, ui);
		}
		else {
			bindNodeBrowserEvents(node, ui);
		}

		return ui;
	}

	// Build and get svg for specified edge.
	// Triggered only once per link (while creating).
	// @return [HTML object] Created HTML element(s).
	var getEdgeSvg = function(edge) {
		var ui = Viva.Graph.svg('g');

		var hover = Viva.Graph.svg('polyline')
			.attr('fill', 'none')
			.attr('stroke', 'transparent') // get the hover efect easier
			.attr('stroke-width', 20);

		var line = Viva.Graph.svg('polyline')
			.attr('fill', 'none')
			.attr('stroke', self.entity.normal.borderColor)
			.attr('stroke-width', 1);

		var icon = Viva.Graph.svg('image')
			.attr('width', self.entity.icon.width)
			.attr('height', self.entity.icon.height)
			.attr('x', 0)
			.attr('y', 0);
		
		if(edge.data.popup != undefined)
			$(ui).popup({ msg: edge.data.popup });

		ui.append(hover);
		ui.append(line);
		ui.append(icon);

		if(!self.isBrowser) {
			bindEdgeEvents(edge, ui);
		}

		return ui;
	}

	var renderNode = function(nodeUI, pos) {
		var x = pos.x - self.entity.width / 2;
        var y = pos.y - self.entity.height / 2;

        if(nodeUI.attr('selected') == 'true') {
        	// Marked by selection
        	nodeUI.childNodes[0]
				.attr('fill', self.entity.hover.backgroundColor)
				.attr('stroke', self.entity.hover.borderColor);
        }
        else if(nodeUI.attr('highlighted') == 'true') {
        	// Hover element
        	nodeUI.childNodes[0]
				.attr('fill', self.entity.normal.backgroundColor)
				.attr('stroke', self.entity.hover.borderColor);
        }
        else {
        	// Others...
        	nodeUI.childNodes[0]
				.attr('fill', self.entity.normal.backgroundColor)
				.attr('stroke', self.entity.normal.borderColor);
        }

        // Make unimportant nodes half-visible
        // when any node is hovered
        if(nodeUI.attr('highlighted') == 'false'
        && nodeUI.attr('selected') != 'true') {
        	nodeUI.attr('opacity', '0.4');
        }
        else {
        	nodeUI.attr('opacity', '1.0');
        }

        // Dashed border for deleted nodes
        if(nodeUI.node.data.deleted === true) {
        	nodeUI.attr('stroke-dasharray', '4,3');
        }
        else {
        	nodeUI.attr('stroke-dasharray', '0');
        }

        nodeUI.attr('transform', 'translate('+ x +','+ y +')');
	}

	var renderEdge = function(linkUI, fromPos, toPos) {
	    // polylines attributes
		var anchors = getEdgePoints(fromPos, toPos);

		var points = '';
		anchors.forEach(function(anchor) {
			points += anchor.x +','+ anchor.y +' ';
		});

	    // image attributes
		var A = getNodePosDetails(fromPos, toPos);
		var angle = 0;

		var x = anchors[0].x - self.entity.icon.width / 2;
		var y = anchors[0].y - self.entity.icon.height / 2;
		switch(A.dir) {
			case 'Top':    y += self.entity.icon.height / 2; break;
			case 'Bottom': y -= self.entity.icon.height / 2; break;
			case 'Left':   x += self.entity.icon.width / 2; break;
			case 'Right':  x -= self.entity.icon.width / 2; break;
		}

		// update layout
	    linkUI.childNodes[0].attr('points', points);
	    linkUI.childNodes[1].attr('points', points);
	    linkUI.childNodes[2].link(self.entity.icon.edge[A.dir])
	    	.attr('x', x)
	    	.attr('y', y);
	}

	var zoomIn = function() {
		self.renderer.zoomIn();
	}

	var zoomOut = function() {
		self.renderer.zoomIn();
	}

	var getNodePosDetails = function(posA, posB) {
		// quarter sign
		var quarter = {x: 1, y: 1};
		if(posA.x >= posB.x) quarter.x = -1;
		if(posA.y >= posB.y) quarter.y = -1;

		// intersection with the rectangle
		var cross = {x: 0, y: 0};
		var dir = '';

		var a = (posB.y - posA.y)/(posB.x - posA.x);
		var dx = quarter.x * self.entity.width / 2;
		var dy = quarter.y * self.entity.height / 2;

		var fy = a * dx;
		var fx = dy / a;
		if(Math.abs(fy) < self.entity.height / 2) {
			cross = {x: quarter.x * self.entity.width / 2, y: fy};
			dir = quarter.x > 0 ? 'Left' : 'Right';
		}
		else {
			cross = {x: fx, y: quarter.y * self.entity.height / 2};
			dir = quarter.y > 0 ? 'Top' : 'Bottom';
		}

		return {
			pos: posA,
			dir: dir, // prefered direction
			cross: cross, // the point at which the straight line intersects the rectangle
			quarter: quarter
		};
	}

	// Calculate midpoints.
	// @param A, B First/Second position details (@see: getNodePosDetails).
	var getEdgeMidpoints = function(A, B) {
		var fixedPosA = {
			x: A.pos.x + A.cross.x,
			y: A.pos.y + A.cross.y
		};
		var fixedPosB = {
			x: B.pos.x + B.cross.x,
			y: B.pos.y + B.cross.y
		};

		var points = [fixedPosA];
		if(A.dir == 'Top' || A.dir == 'Bottom') {
			var m_y = (A.pos.y + B.pos.y) / 2;
			points.push({x: fixedPosA.x, y: m_y});
			points.push({x: fixedPosB.x, y: m_y});
		}
		else {
			var m_x = (A.pos.x + B.pos.x) / 2;
			points.push({x: m_x, y: fixedPosA.y});
			points.push({x: m_x, y: fixedPosB.y});
		}

		points.push(fixedPosB);
		return points;
	}

	var getEdgePoints = function(posA, posB) {
		var A = getNodePosDetails(posA, posB);
		var B = getNodePosDetails(posB, posA);

		if(Math.abs(B.pos.x - A.pos.x) < self.entity.width*3/4) {
			var rx = (B.pos.x + A.pos.x)/2;
			return [
				{x: rx, y: A.pos.y + A.cross.y},
				{x: rx, y: B.pos.y + B.cross.y}
			];
		}

		if(Math.abs(B.pos.y - A.pos.y) < self.entity.height/2) {
			var ry = (B.pos.y + A.pos.y)/2;
			return [
				{x: A.pos.x + A.cross.x, y: ry},
				{x: B.pos.x + B.cross.x, y: ry}
			];
		}

		return getEdgeMidpoints(A, B);
	}

	var getGraphSvg = function() {
		return self.element.html();
	}

	initialize(domPath, nodes, edges, browser);

	// Public methods
	return {
		zoomIn: zoomIn,
		zoomOut: zoomOut,

		getSvg: getGraphSvg,
		notifyNodePropertyChange: notifyNodePropertyChange
	};
};